#!/usr/bin/env python3

import gzip
from probs_runner import probs_enhance_data
from utils import make_observation


def test_output_in_nt_format(tmp_path, script_source_dir):
    original_filename = tmp_path / "original.ttl"
    enhanced_filename = tmp_path / "enhanced.nt.gz"
    with open(original_filename, "wt") as f:
        f.write(make_observation(object=":Cake", measurement=5.1) + "\n")
        f.write(":Cake :objectEquivalentTo :Torta .\n")

    probs_enhance_data(
        original_filename,
        enhanced_filename,
        tmp_path / "working_enhanced",
        script_source_dir,
    )

    with gzip.open(enhanced_filename, "rt") as f:
        lines = f.readlines()

    # Basic check for NT format (no "@prefix")
    assert lines[0].startswith("<")



# -*- coding: utf-8 -*-
"""Test case of an object composed of objects, composed of objects, ... with
observations only in the bottom level of this hierarchy."""

import pytest
from random import randrange, random

from utils import make_observation


OBJECT_FACTS = r"""
:Object0 :objectComposedOf  :Object0.0 ,
                            :Object0.1 ,
                            :Object0.2 .

:Object0.0 :objectComposedOf    :Object0.0.0 ,
                                :Object0.0.1 ,
                                :Object0.0.2 .
:Object0.1 :objectComposedOf    :Object0.1.0 ,
                                :Object0.1.1 ,
                                :Object0.1.2 .
:Object0.2 :objectComposedOf    :Object0.2.0 ,
                                :Object0.2.1 ,
                                :Object0.2.2 .

:Object0.0.0 :objectComposedOf  :Object0.0.0.0 ,
                                :Object0.0.0.1 ,
                                :Object0.0.0.2 .
:Object0.0.1 :objectComposedOf  :Object0.0.1.0 ,
                                :Object0.0.1.1 ,
                                :Object0.0.1.2 .
:Object0.0.2 :objectComposedOf  :Object0.0.2.0 ,
                                :Object0.0.2.1 ,
                                :Object0.0.2.2 .
:Object0.1.0 :objectComposedOf  :Object0.1.0.0 ,
                                :Object0.1.0.1 ,
                                :Object0.1.0.2 .
:Object0.1.1 :objectComposedOf  :Object0.1.1.0 ,
                                :Object0.1.1.1 ,
                                :Object0.1.1.2 .
:Object0.1.2 :objectComposedOf  :Object0.1.2.0 ,
                                :Object0.1.2.1 ,
                                :Object0.1.2.2 .
:Object0.2.0 :objectComposedOf  :Object0.2.0.0 ,
                                :Object0.2.0.1 ,
                                :Object0.2.0.2 .
:Object0.2.1 :objectComposedOf  :Object0.2.1.0 ,
                                :Object0.2.1.1 ,
                                :Object0.2.1.2 .
:Object0.2.2 :objectComposedOf  :Object0.2.2.0 ,
                                :Object0.2.2.1 ,
                                :Object0.2.2.2 .
"""


def test_obs_obj1_comp3_obs2_min_inferred_obs_6(probs_count_inferred_obs):
    """
    1 Object
    3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 6
    User time (seconds): ~1E0
    """

    objects = [f":Object{i}.{j}" for i in range(1) for j in range(3)]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 6


def test_obs_obj1_comp3_obs2_max_inferred_obs_8(probs_count_inferred_obs):
    """
    1 Object
    3 components
    2 Observations for each object at the bottom
    upper bound (compatible observations)
    Observations to be inferred: 8
    User time (seconds): ~1E0
    """

    objects = [f":Object{i}.{j}" for i in range(1) for j in range(3)]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 8


def test_obs_obj1_comp3x3_obs2_min_inferred_obs_18(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 18
    User time (seconds): ~1E0
    """

    objects = [
        f":Object{i}.{j}.{k}" for i in range(1) for j in range(3) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 18


def test_obs_obj1_comp3x3_obs2_max_inferred_obs_512(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components
    2 Observations for each object at the bottom
    upper bound (compatible observations)
    Observations to be inferred: 512
    User time (seconds): ~1E0
    """

    objects = [
        f":Object{i}.{j}.{k}" for i in range(1) for j in range(3) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 512


def test_obs_obj1_comp3x3x3_obs2_min_inferred_obs_54(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 54
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(3) for j in range(3) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 54


@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_obj1_comp3x3x3_obs2_max_inferred_obs_134217728(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components
    2 Observations for each object at the bottom
    upper bound (compatible observations)
    Observations to be inferred: 134217728
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(3) for j in range(3) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 134217728



# -*- coding: utf-8 -*-

from decimal import Decimal
from math import isnan
from pathlib import Path
import pytest
import csv
from io import StringIO

from probs_runner import load_datasource



class TestCOMTRADECodes105:
    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_enhanced):
        with probs_endpoint_from_enhanced() as rdfox:
            yield rdfox

    # @pytest.mark.xfail(reason="extra unnamed parent object is created")
    def test_expected_number_of_objects(self, rdfox):
        results = rdfox.query_records(r"""
            SELECT ?Object ?ObjectName ?ParentName
            WHERE {
                ?Object a :Object .
                OPTIONAL { ?Object rdfs:label ?ObjectName }.
                OPTIONAL { ?Parent :objectComposedOf ?Object ; rdfs:label ?ParentName }.
            }
            ORDER BY ?ObjectName
        """)
        assert len(results) == 6709

    def test_expected_direct_observations(self, rdfox):
        results = rdfox.query_records(r"""
            SELECT ?ObjectName ?Measurement
            WHERE {
                ?Object a :Object ; rdfs:label ?ObjectName .
                FILTER( STRSTARTS(?ObjectName, "COMTRADE Object from Code 105") ||
                        ?ObjectName = "COMTRADE Object from Code 1" ||
                        ?ObjectName = "COMTRADE Object from Code TOTAL" )
                OPTIONAL { ?Observation a :DirectObservation ;
                                  :objectDefinedBy ?Object ;
                                  :measurement ?Measurement . }
            }
            ORDER BY ?ObjectName
        """)
        to_check = [(row["ObjectName"], row["Measurement"]) for row in results]
        assert to_check == [
            ('COMTRADE Object from Code 1', None),
            ('COMTRADE Object from Code 105', 0.0),
            ('COMTRADE Object from Code 10511', 52650005.0),
            ('COMTRADE Object from Code 10512', 0.0),
            ('COMTRADE Object from Code 10513', 5910.0),
            ('COMTRADE Object from Code 10514', None),
            ('COMTRADE Object from Code 10515', None),
            ('COMTRADE Object from Code 10594', 62417.0),
            ('COMTRADE Object from Code 10599', None),
            ('COMTRADE Object from Code TOTAL', None)
        ]

    def test_expected_number_of_inferred_observations(self, rdfox):
        results = rdfox.query_records(r"""
            SELECT ?Observation ?ObjectName ?Measurement
            WHERE {
                ?Observation a :InferredObservation ;
                     :objectDefinedBy [ rdfs:label ?ObjectName ] ;
                     :measurement ?Measurement
            }
            ORDER BY ?ObjectName ?Measurement
        """)
        to_check = [(row["ObjectName"], int(row["Measurement"])) for row in results]
        assert to_check == [
            ('COMTRADE Object from Code 1', 0),              # 2a. inferred from direct obs on 105
            ('COMTRADE Object from Code 1', 52718332),       # 2b. inferred from (1)
            ('COMTRADE Object from Code 105', 52718332),     # 1.  sum of direct, lower bound
            ('COMTRADE Object from Code TOTAL', 0),          # 3a. inferred from (2a)
            ('COMTRADE Object from Code TOTAL', 52718332),   # 3b. inferred from (2b)

        ]



# -*- coding: utf-8 -*-
"""Test case of an object composed of objects, composed of objects, ... with
observations at different level of this hierarchy."""

import pytest
from random import random
from utils import make_observation

levels = 6

OBJECT_FACTS = \
    '\n'.join([f":Object0 :objectComposedOf :Object0.{i} ." for i in range(levels)]) + \
        '\n'.join([f":Object0.{i} :objectComposedOf :Object0.{i}.{j} ." for i in range(levels) for j in range(levels)]) + \
            '\n'.join([f":Object0.{i}.{j} :objectComposedOf :Object0.{i}.{j}.{k} ." for i in range(levels) for j in range(levels) for k in range(levels)]) + \
                '\n'.join([f":Object0.{i}.{j}.{k} :objectComposedOf :Object0.{i}.{j}.{k}.{l} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels)]) + \
                    '\n'.join([f":Object0.{i}.{j}.{k}.{l} :objectComposedOf :Object0.{i}.{j}.{k}.{l}.{m} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels) for m in range(levels)]) + \
                        '\n'.join([f":Object0.{i}.{j}.{k}.{l}.{m} :objectComposedOf :Object0.{i}.{j}.{k}.{l}.{m}.{n} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels) for m in range(levels) for n in range(levels)])

def test_obs_obj1_comp3x3x3x3x3_obs2x2_min_inferred_obs_648(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components
    2 Observations for each object for the bottom 2 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 648
    User time (seconds): ~1E0
    """

    components = 3
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)+2*pow(components,4)

def test_obs_obj1_comp3x3x3x3x3_obs2x2x2_min_inferred_obs_702(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components
    2 Observations for each object for the bottom 3 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 702
    User time (seconds): ~1E0
    """

    components = 3
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}" for i in range(components) for j in range(components) for k in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)+2*pow(components,4)+2*pow(components,3)

def test_obs_obj1_comp3x3x3x3x3_obs5x5_min_inferred_obs_1620(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components
    5 Observations for each object for the bottom 2 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 1620
    User time (seconds): ~1E0
    """

    components = 3
    obs_labels = "abcde"
    n_obs = len(obs_labels)
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in obs_labels
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == n_obs*pow(components,5)+n_obs*pow(components,4)

def test_obs_obj1_comp3x3x3x3x3_obs5x5x5_min_inferred_obs_1755(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components
    5 Observations for each object for the bottom 3 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 1755
    User time (seconds): ~1E0
    """

    components = 3
    obs_labels = "abcde"
    n_obs = len(obs_labels)
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}" for i in range(components) for j in range(components) for k in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in obs_labels
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == n_obs*pow(components,5)+n_obs*pow(components,4)+n_obs*pow(components,3)

@pytest.mark.slow
def test_obs_obj1_comp5x5x5x5x5_obs5x5_min_inferred_obs_18750(probs_count_inferred_obs):
    """
    1 Object
    5 components each with 5 components each with 5 components each with 5 components each with 5 components
    5 Observations for each object for the bottom 2 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 18750
    User time (seconds): ~1E1
    """

    components = 5
    obs_labels = "abcde"
    n_obs = len(obs_labels)
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in obs_labels
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == n_obs*pow(components,5)+n_obs*pow(components,4)

@pytest.mark.slow
def test_obs_obj1_comp5x5x5x5x5_obs5x5x5_min_inferred_obs_19375(probs_count_inferred_obs):
    """
    1 Object
    5 components each with 5 components each with 5 components each with 5 components each with 5 components
    5 Observations for each object for the bottom 3 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 19375
    User time (seconds): ~1E1
    """

    components = 5
    obs_labels = "abcde"
    n_obs = len(obs_labels)
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}" for i in range(components) for j in range(components) for k in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in obs_labels
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == n_obs*pow(components,5)+n_obs*pow(components,4)+n_obs*pow(components,3)

@pytest.mark.slow
def test_obs_obj1_comp5x5x5x5x5_obs5x5x5x5_min_inferred_obs_19500(probs_count_inferred_obs):
    """
    1 Object
    5 components each with 5 components each with 5 components each with 5 components each with 5 components
    5 Observations for each object for the bottom 4 levels
    lower bound (non-compatible observations)
    Observations to be inferred: 19500
    User time (seconds): ~1E1
    """

    components = 5
    obs_labels = "abcde"
    n_obs = len(obs_labels)
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ] + [
        f":Object0.{i}.{j}.{k}" for i in range(components) for j in range(components) for k in range(components)
    ] + [
        f":Object0.{i}.{j}" for i in range(components) for j in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in obs_labels
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == n_obs*pow(components,5)+n_obs*pow(components,4)+n_obs*pow(components,3)+n_obs*pow(components,2)



# -*- coding: utf-8 -*-
"""
"""

from decimal import Decimal

from probs_runner import PROBS

from rdflib import Namespace

QUANTITYKIND = Namespace("http://qudt.org/vocab/quantitykind/")


FACTS = r"""
    PREFIX quantitykind:        <http://qudt.org/vocab/quantitykind/>

    :CementInCOMTRADE :objectEquivalentTo :Cement .
    :CementInPRODCOM :objectEquivalentTo :Cement .

    :Obs1 :objectDirectlyDefinedBy :CementInCOMTRADE .
    :Obs2 :objectInferredDefinedBy :CementInCOMTRADE .

    :Obs1 :hasRole :Role0 ; :hasTime :TP0 ; :hasRegion :R0 ; :hasMetric quantitykind:Mass ; :hasBound :ExactBound .
    :Obs2 :hasRole :Role0 ; :hasTime :TP0 ; :hasRegion :R0 ; :hasMetric quantitykind:Mass ; :hasBound :ExactBound .

    :CementInCOMTRADE :objectComposedOf :Object1 .
    # :CementInCOMTRADE :objectComposedOf :Object2 .
    :CementInCOMTRADE :objectComposedOf :Object3 .
    :Object3 :objectComposedOf :Object4 .
    :Object5 :objectComposedOf :Object4 .
    :Object5 :objectComposedOf :Object6 .

    :Obs3 :objectDirectlyDefinedBy :Object1 .
    :Obs4 :objectDirectlyDefinedBy :Object1 .
    # :Obs4 :objectInferredDefinedBy :Object1 .
    :Obs5 :objectDirectlyDefinedBy :Object3 .
    :Obs44 :objectDirectlyDefinedBy :Object4 .
    :Obs66 :objectDirectlyDefinedBy :Object6 .

    :Obs3 :hasRole :Role1 ; :hasTime :TP1 ; :hasRegion :R1 ; :hasMetric quantitykind:Mass .
    :Obs4 :hasRole :Role2 ; :hasTime :TP1 ; :hasRegion :R1 ; :hasMetric quantitykind:Mass .
    :Obs5 :hasRole :Role1 ; :hasTime :TP1 ; :hasRegion :R1 ; :hasMetric quantitykind:Mass .
    :Obs44 :hasRole :Role1 ; :hasTime :TP1 ; :hasRegion :R1 ; :hasMetric quantitykind:Mass .
    :Obs66 :hasRole :Role3 ; :hasTime :TP1 ; :hasRegion :R1 ; :hasMetric quantitykind:Mass .

    :Obs3 :measurement 10 .
    :Obs4 :measurement 100 .
    :Obs5 :measurement 1000 .
    :Obs44 :measurement 2 .
    :Obs66 :measurement 3 .

    :Obs3 :hasBound :ExactBound .
    :Obs4 :hasBound :ExactBound .
    :Obs5 :hasBound :ExactBound .
    :Obs44 :hasBound :ExactBound .
    :Obs66 :hasBound :ExactBound .
"""


QUERY_INF_ODB = r"""
    SELECT ?Object ?Observation
    WHERE {
        ?Observation :objectDefinedBy ?Object ;
             :objectInferredDefinedBy ?Object .
    }
    ORDER BY ?Observation ?Object
"""

QUERY_BOUNDS = r"""
    SELECT ?Observation ?Bound ?Measurement
    WHERE {
        ?Observation :measurement ?Measurement ;
             :hasBound ?Bound .
    }
    ORDER BY ?Observation ?Bound ?Measurement
"""

QUERY_WDF = r"""
    PREFIX prov: <http://www.w3.org/ns/prov#>
    SELECT ?Observation ?WDF ?Measurement
    WHERE {
        ?Observation :measurement ?Measurement ;
             prov:wasDerivedFrom ?WDF .
    }
    ORDER BY ?Observation ?WDF ?Measurement
"""

QUERY_PROPS = r"""
    SELECT ?Observation ?Role ?Time ?Region ?Metric ?Measurement
    WHERE {
        ?Observation :hasRole ?Role ;
             :hasTime ?Time ;
             :hasRegion ?Region ;
             :hasMetric ?Metric ;
             :measurement ?Measurement .
    }
    ORDER BY ?Observation ?Role ?Time ?Region ?Metric ?Measurement
"""

def test_inferred_odb(probs_answer_query):

    results = probs_answer_query(FACTS, QUERY_INF_ODB)

    assert len(results) == 17

    to_check = set([(row["Object"]) for row in results])
    # note that `set()` removes all duplicates
    assert to_check == set([
            # Obs1
            (PROBS.Cement),
            (PROBS.CementInPRODCOM),
            # Obs2
            (PROBS.Cement),
            (PROBS.CementInCOMTRADE),
            (PROBS.CementInPRODCOM),
            # Obs3 - Obs5
            (PROBS.Cement),
            (PROBS.CementInCOMTRADE),
            (PROBS.CementInPRODCOM),
            # Obs4 (lower-bound)
            (PROBS.Cement),
            (PROBS.CementInCOMTRADE),
            (PROBS.CementInPRODCOM),
            # Obs44
            (PROBS.Object3),
            (PROBS.Object5),
            # Obs66
            (PROBS.Object5),
            # Obs3 - inferred from Obs44
            (PROBS.Cement),
            (PROBS.CementInCOMTRADE),
            (PROBS.CementInPRODCOM),
    ])


def test_bounds(probs_answer_query):

    results = probs_answer_query(FACTS, QUERY_BOUNDS)

    assert len(results) == 11

    to_check = set([(row["Bound"], row["Measurement"]) for row in results])
    # note that `set()` removes all duplicates
    assert to_check == set([
            # Obs3
            (PROBS.ExactBound, Decimal("10")),
            # Obs4
            (PROBS.ExactBound, Decimal("100")),
            # Obs5
            (PROBS.ExactBound, Decimal("1000")),
            # Obs44
            (PROBS.ExactBound, Decimal("2")),
            # Obs44 -> Object3
            (PROBS.ExactBound, Decimal("2")),
            # Obs44 -> Object5
            (PROBS.LowerBound, Decimal("2")),
            # Obs66
            (PROBS.ExactBound, Decimal("3")),
            # Obs66 -> Object5
            (PROBS.LowerBound, Decimal("3")),
            # Obs3 - Obs5
            (PROBS.ExactBound, Decimal("1010")),
            # Obs4 (lower-bound)
            (PROBS.LowerBound, Decimal("100")),
            # Obs3 - inferred from Obs44
            (PROBS.ExactBound, Decimal("12")),
    ]) # note that those of Obs1 and Obs2 are not included here because they have no measurement (and the query asks explicitly about the measurement)


def test_wdf(probs_answer_query):

    results = probs_answer_query(FACTS, QUERY_WDF)

    assert len(results) == 11

    to_check = set([(row["WDF"], row["Measurement"]) for row in results])
    # note that `set()` removes all duplicates
    # we cannot check all of them because some IRIs are automatically generated (and so we need to use `issuperset`)
    assert to_check.issuperset(set([
            # Obs44 -> Object3
            (PROBS.Obs44, Decimal("2")),
            # Obs44 -> Object5
            (PROBS.Obs44, Decimal("2")),
            # Obs66 -> Object5
            (PROBS.Obs66, Decimal("3")),
            # Obs3 - Obs5 -> Cement in COMTRADE
            (PROBS.Obs3, Decimal("1010")),
            (PROBS.Obs5, Decimal("1010")),
            # Obs4 (lower-bound) -> Cement in COMTRADE
            (PROBS.Obs4, Decimal("100")),
            # Obs3 - inferred from "Obs44 -> Object3" -> Cement in COMTRADE
            (PROBS.Obs3, Decimal("12")),
            # (PROBS.Obs44, Decimal("12")), #we now include only the directly connected ones
    ])) # note that those of Obs1 and Obs2 are not included here because they have no measurement (and the query asks explicitly about the measurement)

    # there are 3 additional ones due to Equivalence (we partially check them below)
    derived_from_self = [
        True
        for row in results
        if row["Observation"] == row["WDF"]
    ]
    assert len(derived_from_self) == 3

    # and one for "inferred from Obs44 -> Object3" -> Cement in COMTRADE (we partially check it below),
    measurement_equals_12 = [
        True
        for row in results
        if row["Measurement"] == Decimal("12")
    ]
    assert len(measurement_equals_12) == 3


def test_props(probs_answer_query):

    results = probs_answer_query(FACTS, QUERY_PROPS)

    assert len(results) == 11

    to_check = set([(row["Role"], row["Time"], row["Region"], row["Metric"], row["Measurement"]) for row in results])
    # note that `set()` removes all duplicates
    assert to_check == set([
            # Obs3
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("10")),
            # Obs4 or Obs4 (lower-bound)
            (PROBS.Role2, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("100")),
            # Obs5
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("1000")),
            # Obs44 or Obs44 -> Object3
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("2")),
            # Obs44 or Obs44 -> Object5
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("2")),
            # Obs66 or Obs66 -> Object5
            (PROBS.Role3, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("3")),
            # Obs66 or Obs66 -> Object5
            (PROBS.Role3, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("3")),
            # Obs3 - Obs5
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("1010")),
            # Obs3 - inferred from Obs44
            (PROBS.Role1, PROBS.TP1, PROBS.R1, QUANTITYKIND.Mass, Decimal("12")),
    ])



# -*- coding: utf-8 -*-
"""When an Object is composed of other objects, the measurements of compatible
objects should be summed.

"""

from decimal import Decimal
import pytest

from probs_runner import PROBS

from utils import make_observation


BAKED_GOODS_FACTS = r"""
    :Cake a :Object .
    :Bread a :Object .
    :BakedGoods a :Object .
    :BakedGoods :objectComposedOf :Cake, :Bread .
"""


QUERY = r"""
    SELECT ?observation ?role ?time ?region ?metric ?measurement ?bound
    WHERE {
        ?observation :objectDefinedBy :BakedGoods ;
        :hasRole ?role ;
        :hasTime ?time ;
        :hasRegion ?region ;
        :hasMetric ?metric ;
        :measurement ?measurement ;
        :hasBound ?bound .
    }
    ORDER BY ?role ?time ?region ?metric
"""


@pytest.mark.parametrize("role", [":SoldProduction", ":Consumption", ":Import", ":Export"])
def test_two_compatible_observations_are_summed_for_roles_with_only_object(probs_answer_query, role):
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(role=role, object=":Cake", measurement=5.1)
        + make_observation(role=role, object=":Bread", measurement=22.6)
    )
    results = probs_answer_query(facts, QUERY)

    assert len(results) == 1
    result = results[0]
    assert "ComposedInferredObservation-" in result["observation"]
    assert result["role"] == PROBS[role[1:]]  # remove empty : prefix
    assert result["time"] == PROBS.TimePeriod_YearOf2018
    assert result["region"] == PROBS.UnitedKingdom
    assert result["metric"] == "mass"
    assert result["measurement"] == Decimal("27.7")
    assert result["bound"] == PROBS.ExactBound


@pytest.mark.parametrize("role", [":ProcessInput", ":ProcessOutput"])
def test_two_compatible_observations_are_summed_for_process_input_output(probs_answer_query, role):
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(role=role, process=":Bakery", object=":Cake", measurement=5.1)
        + make_observation(role=role, process=":Bakery", object=":Bread", measurement=22.6)
    )
    results = probs_answer_query(facts, QUERY)

    assert len(results) == 1
    result = results[0]
    assert "ComposedInferredObservation-" in result["observation"]
    assert result["role"] == PROBS[role[1:]]  # remove empty : prefix
    assert result["time"] == PROBS.TimePeriod_YearOf2018
    assert result["region"] == PROBS.UnitedKingdom
    assert result["metric"] == "mass"
    assert result["measurement"] == Decimal("27.7")
    assert result["bound"] == PROBS.ExactBound


def test_compatible_sets_of_observations_are_summed_separately(probs_answer_query):
    """Like the previous test, but now with some observations from 2019 too."""
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(
            time=":TimePeriod_YearOf2018", object=":Cake", measurement=5.1
        )
        + make_observation(
            time=":TimePeriod_YearOf2018", object=":Bread", measurement=22.6
        )
        + make_observation(
            time=":TimePeriod_YearOf2019", object=":Cake", measurement=3.1
        )
        + make_observation(
            time=":TimePeriod_YearOf2019", object=":Bread", measurement=20.0
        )
    )
    results = probs_answer_query(facts, QUERY)

    to_check = [(x["time"], x["measurement"], x["bound"]) for x in results]
    assert to_check == [
        (PROBS.TimePeriod_YearOf2018, Decimal("27.7"), PROBS.ExactBound),
        (PROBS.TimePeriod_YearOf2019, Decimal("23.1"), PROBS.ExactBound),
    ]


def test_observations_with_different_roles_are_not_composed(probs_answer_query):
    """Different roles are not compatible -- only lower bounds are inferred for
    BakedGoods."""
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Bread", role=":Import", measurement=3.2)
        + make_observation(object=":Cake", role=":SoldProduction", measurement=5.4)
    )
    results = probs_answer_query(facts, QUERY)

    to_check = [(x["role"], x["measurement"], x["bound"]) for x in results]
    assert to_check == [
        (PROBS.Import, Decimal("3.2"), PROBS.LowerBound),
        (PROBS.SoldProduction, Decimal("5.4"), PROBS.LowerBound),
    ]


def test_observations_with_different_time_periods_are_not_composed(probs_answer_query):
    """Different time periods are not compatible."""
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(
            object=":Cake", time=":TimePeriod_YearOf2018", measurement=3.2
        )
        + make_observation(
            object=":Bread", time=":TimePeriod_YearOf2019", measurement=5.4
        )
    )
    results = probs_answer_query(facts, QUERY)

    to_check = [(x["time"], x["measurement"], x["bound"]) for x in results]
    assert to_check == [
        (PROBS.TimePeriod_YearOf2018, Decimal("3.2"), PROBS.LowerBound),
        (PROBS.TimePeriod_YearOf2019, Decimal("5.4"), PROBS.LowerBound),
    ]


def test_observations_with_different_regions_are_not_composed(probs_answer_query):
    """Different regions are not compatible."""
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Bread", region=":Italy", measurement=3.2)
        + make_observation(object=":Cake", region=":UnitedKingdom", measurement=5.4)
    )
    results = probs_answer_query(facts, QUERY)

    to_check = [(x["region"], x["measurement"], x["bound"]) for x in results]
    assert to_check == [
        (PROBS.Italy, Decimal("3.2"), PROBS.LowerBound),
        (PROBS.UnitedKingdom, Decimal("5.4"), PROBS.LowerBound),
    ]


def test_observations_with_different_metrics_are_not_composed(probs_answer_query):
    """Different metrics are not compatible."""
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", metric='"mass"', measurement=3.2)
        + make_observation(object=":Bread", metric='"volume"', measurement=5.4)
    )
    results = probs_answer_query(facts, QUERY)

    to_check = [(x["metric"], x["measurement"], x["bound"]) for x in results]
    assert to_check == [
        ("mass", Decimal("3.2"), PROBS.LowerBound),
        ("volume", Decimal("5.4"), PROBS.LowerBound),
    ]



# -*- coding: utf-8 -*-
"""When an Object is composed of other objects, inferred observations should be
marked as lower bounds if any observations are missing.

"""

from decimal import Decimal

from probs_runner import PROBS

from utils import make_observation


# These are the sample objects we will test.
BAKED_GOODS_FACTS = r"""
    :Cake a :Object .
    :Bread a :Object .
    :BakedGoods a :Object .
    :BakedGoods :objectComposedOf :Cake, :Bread .
"""

QUERY = r"""
    SELECT ?measurement ?bound
    WHERE {
        ?observation :objectDefinedBy :BakedGoods ;
                :measurement ?measurement ;
                :hasBound ?bound .
    }
    ORDER BY ?measurement ?bound
"""


def test_nothing_is_inferred_with_no_inputs(probs_answer_query):
    """
    This seems a fairly trivial test, but it could allow to test if we generate too many `EmptyObservation`.
    """
    results = probs_answer_query(BAKED_GOODS_FACTS, QUERY)
    assert len(results) == 0


def test_lower_bound_is_inferred_with_missing_observation(probs_answer_query):
    """
    A lower bound should be generated when we have no observation for `Bread`. This inferred observation should have the same `measurement` of the only observation we have for `Cake`.
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=5.1)
        # no observation for bread
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.LowerBound, "measurement": Decimal("5.1")},
    ]


def test_lower_bound_is_inferred_with_missing_measurement(probs_answer_query):
    """
    If an observation does not have a measurement, the output is the same as `test_lower_bound_is_inferred_with_missing_observation`
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=5.1)
        + make_observation(object=":Bread", measurement=None)
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.LowerBound, "measurement": Decimal("5.1")},
    ]


def test_lower_bound_is_inferred_with_only_missing_measurements(probs_answer_query):
    """
    If we combine two observations without measurements, we should get a lower bound with `0` as `measurement`.
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=None)
        + make_observation(object=":Bread", measurement=None)
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.LowerBound, "measurement": Decimal("0")},
    ]


def test_exact_bound_is_inferred_with_zero_measurement(probs_answer_query):
    """
    If the `measurement` is `0`, the behaviour should be the same as for any other value.
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=5.1)
        + make_observation(object=":Bread", measurement=0)
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.ExactBound, "measurement": Decimal("5.1")},
    ]


def test_exact_bound_is_inferred_with_only_zero_measurements(probs_answer_query):
    """
    If the `measurement` is `0`, the behaviour should be the same as for any other value.
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=0)
        + make_observation(object=":Bread", measurement=0)
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.ExactBound, "measurement": Decimal("0")},
    ]


def test_lower_bound_is_inferred_with_zero_and_missing_measurement(probs_answer_query):
    """
    If the `measurement` is `0`, the behaviour should be the same as for any other value. 
    But since we have no measurement for `Bread` here, a lower bound should be derived.
    """
    facts = (
        BAKED_GOODS_FACTS
        + make_observation(object=":Cake", measurement=0)
        + make_observation(object=":Bread", measurement=None)
    )
    results = probs_answer_query(facts, QUERY)
    assert results == [
        {"bound": PROBS.LowerBound, "measurement": Decimal("0")},
    ]



# -*- coding: utf-8 -*-

import pytest
from random import randrange, random


from utils import make_observation


OBJECT_FACTS = r"""
    :Object0 :objectComposedOf  :Object0.0 ,
                                :Object0.1 ,
                                :Object0.2 ,
                                :Object0.3 ,
                                :Object0.4 ,
                                :Object0.5 ,
                                :Object0.6 ,
                                :Object0.7 ,
                                :Object0.8 ,
                                :Object0.9 .
    :Object1 :objectComposedOf  :Object1.0 ,
                                :Object1.1 ,
                                :Object1.2 ,
                                :Object1.3 ,
                                :Object1.4 ,
                                :Object1.5 ,
                                :Object1.6 ,
                                :Object1.7 ,
                                :Object1.8 ,
                                :Object1.9 .
"""


def test_obs_1_10_1each_inferred_obs_1(probs_count_inferred_obs):
    """
    1 Object, 10 components, 1 Observation each
    Observations to be inferred: 1
    User time (seconds): ~1E0
    """
    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=1)
        for i in range(1)
        for j in range(10)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 1


def test_obs_2_10_1each_inferred_obs_2(probs_count_inferred_obs):
    """
    2 Objects, 10 components, 1 Observation each
    Observations to be inferred: 2
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=1)
        for i in range(2)
        for j in range(10)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 2


def test_obs_2_10_2each_inferred_obs_4(probs_count_inferred_obs):
    """
    2 Object, 10 components, 2 Observation each
    Observations to be inferred: 4
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=1, region=f":R{k}")
        for i in range(2)
        for j in range(10)
        for k in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 4


def test_obs_2_10_3each_inferred_obs_6(probs_count_inferred_obs):
    """
    2 Object, 10 components, 3 Observation each
    Observations to be inferred: 6
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=1, region=f":R{k}")
        for i in range(2)
        for j in range(10)
        for k in range(3)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 6


def test_obs_1_10_2eachsame_inferred_obs_1024(probs_count_inferred_obs):
    """
    1 Object, 10 components, 2 Observation each, same probsObs
    Observations to be inferred: 1024
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(1)
        for j in range(10)
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 1024


# @pytest.mark.slow
def test_obs_2_10_2eachsame_inferred_obs_2048(probs_count_inferred_obs):
    """
    2 Object, 10 components, 2 Observation each, same probsObs
    Observations to be inferred: 2048
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(2)
        for j in range(10)
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 2048


@pytest.mark.slow
def test_obs_1_10_3eachsame_inferred_obs_59049(probs_count_inferred_obs):
    """
    1 Object, 10 components, 3 Observations each, same probsObs
    Observations to be inferred: 59049
    User time (seconds): ~1E1
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(1)
        for j in range(10)
        for _ in range(3)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 59049


@pytest.mark.slow
def test_obs_2_10_3eachsame_inferred_obs_118098(probs_count_inferred_obs):
    """
    2 Object, 10 components, 3 Observations each, same probsObs
    Observations to be inferred: 118098
    User time (seconds): ~1E1
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(2)
        for j in range(10)
        for _ in range(3)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 118098


# @pytest.mark.slow
def test_obs_1_5_5eachsame_inferred_obs_3125(probs_count_inferred_obs):
    """
    1 Object, 5 components, 5 Observations each, same probsObs
    Observations to be inferred: 3125
    User time (seconds): ~1E0
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(1)
        for j in range(5)
        for _ in range(5)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 3125


@pytest.mark.slow
def test_obs_1_6_6eachsame_inferred_obs_46656(probs_count_inferred_obs):
    """
    1 Object, 6 components, 6 Observations each, same probsObs
    Observations to be inferred: 46656
    User time (seconds): ~1E1
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(1)
        for j in range(6)
        for _ in range(6)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 46656


@pytest.mark.slow
def test_obs_1_7_6eachsame_inferred_obs_279936(probs_count_inferred_obs):
    """
    1 Object, 7 components, 6 Observations each, same probsObs
    Observations to be inferred: 279936
    User time (seconds): ~1E2
    """

    input_obs = "\n".join(
        make_observation(object=f":Object{i}.{j}", measurement=randrange(100000))
        for i in range(1)
        for j in range(7)
        for _ in range(6)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 279936



# -*- coding: utf-8 -*-

from decimal import Decimal
import pytest

from probs_runner import PROBS
from utils import make_observation

class TestObservationCompositionMultipleSteps:
    """Test longer chains of inferred observations."""

    @pytest.fixture(scope="class")
    def object_facts(self):
        """These are the sample objects we will test."""
        return r"""
        :Cupcakes a :Object .
        :Muffins a :Object .
        :Cake a :Object .
        :Bread a :Object .
        :BakedGoods a :Object .

        :BakedGoods :objectComposedOf :Cake, :Bread .
        :Cake :objectComposedOf :Cupcakes, :Muffins .
        """

    @pytest.fixture(scope="class")
    def all_facts(self, object_facts):
        """Sample data about the objects."""
        return (
            object_facts
            + make_observation(obs_id="1", object=":Cupcakes", measurement=2)
            + make_observation(obs_id="2", object=":Muffins", measurement=3)
            + make_observation(obs_id="3", object=":Bread", measurement=6)
        )

    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_facts, all_facts):
        with probs_endpoint_from_facts(all_facts) as rdfox:
            yield rdfox

    def test_observations_are_summed_across_multiple_steps(self, rdfox):
        results = rdfox.query_records(r"""
            SELECT ?measurement ?bound
            WHERE {
                ?observation :objectDefinedBy :BakedGoods ;
                    :measurement ?measurement ;
                    :hasBound ?bound .
            }
            ORDER BY ?measurement
        """)

        assert results == [
            {"bound": PROBS.ExactBound, "measurement": 11},
        ]

    def test_observation_was_derived_from(self, rdfox):
        # Check the expected observations have been derived with
        # prov:wasDerivedFrom relationships.
        results = rdfox.query_records(r"""
            PREFIX prov: <http://www.w3.org/ns/prov#>
            SELECT ?observation1 ?observation2
            WHERE {
                ?observation1 :objectDefinedBy :Cake ;
                      prov:wasDerivedFrom :Observation1, :Observation2 .
                ?observation2 :objectDefinedBy :BakedGoods ;
                      prov:wasDerivedFrom :Observation3, ?observation1 .
            }
        """)

        assert len(results) == 1

    def test_only_one_exact_observations_at_top_level(self, rdfox):
        results = rdfox.query_records(r"""
            SELECT ?measurement
            WHERE {
                ?observation :objectDefinedBy :BakedGoods ;
                    :measurement ?measurement ;
                    :hasBound :ExactBound .
            }
            ORDER BY ?measurement
        """)

        assert results == [
            {"measurement": 11},
        ]



# -*- coding: utf-8 -*-
"""Test case of an object composed of objects with different numbers of observations."""

import pytest
from random import random
from utils import make_observation

@pytest.mark.slow
def test_ordering_1000_1000_max_top(probs_count_inferred_obs):
    """
    *Input*
    1 Object
    1000 components
    1000 Observations for the first Object
    1 Observation for each other Object
    All compatible observations (max)
    
    *Output*
    Observations to be inferred: 1000
    
    *Stats (1 thread)*
    Lexicographical order:
     - User time (seconds): ~14
     - Maximum resident set size (MB): ~240
    Number of observations order:
     - User time (seconds): ~14
     - Maximum resident set size (MB): ~190
    """

    # Number of components
    num_components = 1000

    # Number of observations of the first component
    num_observations_first = 1000

    run_ordering_test(probs_count_inferred_obs, num_components, num_observations_first)

@pytest.mark.slow
def test_ordering_1000_1000_max_bottom(probs_count_inferred_obs):
    """
    *Input*
    1 Object
    1000 components
    1000 Observations for the first Object
    1 Observation for each other Object
    All compatible observations (max)
    
    *Output*
    Observations to be inferred: 1000
    
    *Stats (1 thread)*
    Lexicographical order:
     - User time (seconds): ~17
     - Maximum resident set size (MB): ~440
    Number of observations order:
     - User time (seconds): ~14
     - Maximum resident set size (MB): ~190
    """

    # Number of components
    num_components = 1000

    # Number of observations of the first component
    num_observations_first = 1000

    run_ordering_test(probs_count_inferred_obs, num_components, num_observations_first, False)

@pytest.mark.slow
def test_ordering_1000_2000_max_top(probs_count_inferred_obs):
    """
    *Input*
    1 Object
    1000 components
    2000 Observations for the first Object
    1 Observation for each other Object
    All compatible observations (max)
    
    *Output*
    Observations to be inferred: 2000

    *Stats (1 thread)*
    Lexicographical order:
     - User time (seconds): ~25
     - Maximum resident set size (MB): ~460
    Number of observations order:
     - User time (seconds): ~25
     - Maximum resident set size (MB): ~360
    """

    # Number of components
    num_components = 1000

    # Number of observations of the first component
    num_observations_first = 2000

    run_ordering_test(probs_count_inferred_obs, num_components, num_observations_first)

@pytest.mark.slow
def test_ordering_1000_2000_max_bottom(probs_count_inferred_obs):
    """
    *Input*
    1 Object
    1000 components
    2000 Observations for the first Object
    1 Observation for each other Object
    All compatible observations (max)
    
    *Output*
    Observations to be inferred: 2000

    *Stats (1 thread)*
    Lexicographical order:
     - User time (seconds): ~34
     - Maximum resident set size (MB): ~850
    Number of observations order:
     - User time (seconds): ~25
     - Maximum resident set size (MB): ~360
    """

    # Number of components
    num_components = 1000

    # Number of observations of the first component
    num_observations_first = 2000

    run_ordering_test(probs_count_inferred_obs, num_components, num_observations_first, False)

def run_ordering_test(probs_count_inferred_obs, num_components, num_observations_first, from_top=True):
    """
    Utility function
    """

    # First from top or bottom
    if from_top:
        num = "1"
    else:
        num = str(num_components)

    # Components
    components = [
        f":Component{i+1}" for i in range(num_components)
    ]

    # Composition hierarchy
    OBJECT_FACTS = '\n'.join(
        [f":Object :objectComposedOf {obj}." for obj in components]
    )

    # 1 observation for each component
    input_obs = "\n".join(
        make_observation(
            object=obj,
            measurement=round(random(), 6),
        )
        for obj in components
    )

    # Additional observations for the first component
    additional_obs = "\n".join(
        make_observation(
            object=":Component" + num,
            measurement=round(random(), 6),
        )
        for _ in range(num_observations_first-1)
    )

    # Check that the number of inferred observations is the same as the number of observations of the first component
    assert probs_count_inferred_obs(OBJECT_FACTS + '\n' + input_obs + '\n' + additional_obs, ":Object") == num_observations_first



# -*- coding: utf-8 -*-

from decimal import Decimal
import pytest

from probs_runner import PROBS


from utils import make_observation
from rdflib import URIRef


# These are the sample objects we will test
SIMPLIFIED_OBJECT_FACTS = r"""
:BGS1 a :Object .
:BGSCrushedStone a :Object .
:CrushedStone a :Object .
:SandGravel a :Object .
:Aggregates a :Object .

:Aggregates :objectComposedOf :CrushedStone, :SandGravel .
:CrushedStone :objectEquivalentTo :BGSCrushedStone .
:BGSCrushedStone :objectComposedOf :BGS1 .
"""


# Template queries for the tests below

OBJECT_QUERY = """
SELECT ?measurement ?bound
WHERE {
    ?observation :objectDefinedBy %s ;
         :measurement ?measurement .
    OPTIONAL { ?observation :hasBound ?bound . }
}
ORDER BY ?measurement
"""

OBJECT_QUERY_TIME = """
SELECT ?measurement ?bound ?year
WHERE {
    ?observation :objectDefinedBy %s ;
         :measurement ?measurement .
    OPTIONAL { ?observation :hasTime ?year . }
    OPTIONAL { ?observation :hasBound ?bound . }
}
ORDER BY ?year ?measurement
"""


class TestSimplifiedBGSExample:
    """Test longer chains of inferred observations including both composition and equivalence.

    The issue arises when A = B + C

    but B has two observations, one direct and one inferred via composition AND equivalence
    """

    @pytest.fixture(scope="class")
    def all_facts(self):
        """Sample data about the objects."""
        return (
            SIMPLIFIED_OBJECT_FACTS
            + make_observation(object=":BGS1", measurement=101000000000.0)  # Obs3
            + make_observation(object=":CrushedStone", measurement=99000000000.0)  # Obs 2
            + make_observation(object=":SandGravel", measurement=56000000000.0)  # Obs 4
        )

    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_facts, all_facts):
        with probs_endpoint_from_facts(all_facts) as rdfox:
            yield rdfox

    def test_observations_crushed_stone(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":CrushedStone")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('99000000000')},  # Obs 2 (direct)
            {"bound": PROBS.ExactBound, "measurement": Decimal('101000000000')}, # Obs 3 (inferred)
        ]

    def test_observations_sand_gravel(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":SandGravel")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('56000000000')},  # Obs 4
        ]


    def test_observations_aggregates(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":Aggregates")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('155000000000')}, # Obs 2+4
            {"bound": PROBS.ExactBound, "measurement": Decimal('157000000000')}, # Obs 3+4
        ]


class TestSimplifiedBGSExampleTwoYears:
    """Test longer chains of inferred observations including both composition and equivalence.

    The issue arises when A = B + C

    but B has two observations, one direct and one inferred via composition AND equivalence
    """

    @pytest.fixture(scope="class")
    def all_facts(self):
        """Sample data about the objects."""
        return (
            SIMPLIFIED_OBJECT_FACTS
            + make_observation(object=":BGS1", time=":TimePeriod_YearOf2014", measurement=101000000000.0)  # Obs3
            + make_observation(object=":CrushedStone", time=":TimePeriod_YearOf2014", measurement=99000000000.0)  # Obs 2
            + make_observation(object=":BGS1", time=":TimePeriod_YearOf2018", measurement=69722551000.0)  # Obs 1
            + make_observation(object=":SandGravel", time=":TimePeriod_YearOf2014", measurement=56000000000.0)  # Obs 4
            + make_observation(object=":SandGravel", time=":TimePeriod_YearOf2018", measurement=56660203000.0)  # Obs 6
        )

    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_facts, all_facts):
        with probs_endpoint_from_facts(all_facts) as rdfox:
            yield rdfox

    def test_observations_crushed_stone(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY_TIME % ":CrushedStone")
        assert results == [
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('99000000000')},  # Obs 2 (direct)
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('101000000000')}, # Obs 3 (inferred)
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2018, "measurement": Decimal('69722551000')}, # Obs 1 (direct)
        ]

    def test_observations_sand_gravel(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY_TIME % ":SandGravel")
        assert results == [
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('56000000000')},  # Obs 4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2018, "measurement": Decimal('56660203000')},  # Obs 6
        ]


    def test_observations_aggregates(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY_TIME % ":Aggregates")
        assert results == [
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('155000000000')}, # Obs 2+4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('157000000000')}, # Obs 3+4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2018, "measurement": Decimal('126382754000')}, # Obs 1+6
        ]


FULL_OBJECT_FACTS = r"""
:BGS1 a :Object .
:BGS2 a :Object .
:BGS3 a :Object .
:Prodcom1 a :Object .
:Prodcom2 a :Object .
:Prodcom3 a :Object .
:Prodcom4 a :Object .
:Prodcom5 a :Object .
:BGSCrushedStone a :Object .
:ProdcomCrushedStone a :Object .
:BGSSandGravel a :Object .
:ProdcomSandGravel a :Object .
:CrushedStone a :Object .
:SandGravel a :Object .
:Aggregates a :Object .

:Aggregates :objectComposedOf :CrushedStone, :SandGravel .
:CrushedStone :objectEquivalentTo :BGSCrushedStone, :ProdcomCrushedStone .
:SandGravel :objectEquivalentTo :BGSSandGravel, :ProdcomSandGravel .
:BGSCrushedStone :objectComposedOf :BGS1, :BGS2, :BGS3 .
:ProdcomCrushedStone :objectComposedOf :Prodcom1, :Prodcom2, :Prodcom3 .
:ProdcomSandGravel :objectComposedOf :Prodcom4, :Prodcom5 .
"""


class TestBGSExampleIncomplete2018:
    """More complex test based on BGS data, with incomplete 2018 observations leading to a lower bound."""

    @pytest.fixture(scope="class")
    def all_facts(self):
        """Sample data about the objects."""
        return (
            FULL_OBJECT_FACTS
            + make_observation(object=":BGS1", time=":TimePeriod_YearOf2014", measurement=38000000000.0)  # ObsA |
            + make_observation(object=":BGS2", time=":TimePeriod_YearOf2014", measurement=53000000000.0)  # ObsB |-> Obs3
            + make_observation(object=":BGS3", time=":TimePeriod_YearOf2014", measurement=10000000000.0)  # ObsC |
            + make_observation(object=":BGSCrushedStone", time=":TimePeriod_YearOf2014", measurement=99000000000.0)  # Obs 2
            + make_observation(object=":BGSSandGravel", time=":TimePeriod_YearOf2014", measurement=56000000000.0)  # Obs 4
            + make_observation(object=":Prodcom1", time=":TimePeriod_YearOf2018", measurement=68391115000.0)  # ObsX |
            + make_observation(object=":Prodcom2", time=":TimePeriod_YearOf2018", measurement=0.0)            # ObsY |-> Obs 1
            + make_observation(object=":Prodcom3", time=":TimePeriod_YearOf2018", measurement=1331436000.0)   # ObsZ |
            + make_observation(object=":Prodcom4", time=":TimePeriod_YearOf2014", measurement=26817475000.0) # | Obs 5
            + make_observation(object=":Prodcom5", time=":TimePeriod_YearOf2014", measurement=24192579000.0) # |
        )

    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_facts, all_facts):
        with probs_endpoint_from_facts(all_facts) as rdfox:
            yield rdfox

    def test_observations_crushed_stone(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":CrushedStone")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('69722551000')},  # Obs 1
            {"bound": PROBS.ExactBound, "measurement": Decimal('99000000000')},  # Obs 2
            {"bound": PROBS.ExactBound, "measurement": Decimal('101000000000')}, # Obs 3 (A+B+C)
        ]

    def test_observations_sand_gravel(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":SandGravel")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('51010054000')},  # Obs 5
            {"bound": PROBS.ExactBound, "measurement": Decimal('56000000000')},  # Obs 4
        ]

    def test_observations_aggregates(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY_TIME % ":Aggregates")
        assert results == [
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('150010054000')},  # Obs 2+5
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('152010054000')},  # Obs 3+5
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('155000000000')},  # Obs 2+4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('157000000000')},  # Obs 3+4
            {"bound": PROBS.LowerBound, "year": PROBS.TimePeriod_YearOf2018, "measurement": Decimal('69722551000')},  # Obs 1 (incomplete)
        ]


class TestBGSExampleComplete2018:
    """More complex test based on BGS data, with complete observations for 2018."""

    @pytest.fixture(scope="class")
    def all_facts(self):
        """Sample data about the objects."""
        return (
            FULL_OBJECT_FACTS
            + make_observation(object=":BGS1", time=":TimePeriod_YearOf2014", measurement=38000000000.0)  # ObsA |
            + make_observation(object=":BGS2", time=":TimePeriod_YearOf2014", measurement=53000000000.0)  # ObsB |-> Obs3
            + make_observation(object=":BGS3", time=":TimePeriod_YearOf2014", measurement=10000000000.0)  # ObsC |
            + make_observation(object=":BGSCrushedStone", time=":TimePeriod_YearOf2014", measurement=99000000000.0)  # Obs 2
            + make_observation(object=":BGSSandGravel", time=":TimePeriod_YearOf2014", measurement=56000000000.0)  # Obs 4
            + make_observation(object=":Prodcom1", time=":TimePeriod_YearOf2018", measurement=68391115000.0)  # ObsX |
            + make_observation(object=":Prodcom2", time=":TimePeriod_YearOf2018", measurement=0.0)            # ObsY |-> Obs 1
            + make_observation(object=":Prodcom3", time=":TimePeriod_YearOf2018", measurement=1331436000.0)   # ObsZ |
            + make_observation(object=":Prodcom4", time=":TimePeriod_YearOf2018", measurement=27047315000.0) # | Obs 6
            + make_observation(object=":Prodcom5", time=":TimePeriod_YearOf2018", measurement=29612888000.0) # |
            + make_observation(object=":Prodcom4", time=":TimePeriod_YearOf2014", measurement=26817475000.0) # | Obs 5
            + make_observation(object=":Prodcom5", time=":TimePeriod_YearOf2014", measurement=24192579000.0) # |
        )

    @pytest.fixture(scope="class")
    def rdfox(self, probs_endpoint_from_facts, all_facts):
        with probs_endpoint_from_facts(all_facts) as rdfox:
            yield rdfox

    def test_observations_crushed_stone(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":CrushedStone")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('69722551000')},  # Obs 1
            {"bound": PROBS.ExactBound, "measurement": Decimal('99000000000')},  # Obs 2
            {"bound": PROBS.ExactBound, "measurement": Decimal('101000000000')}, # Obs 3 (A+B+C)
        ]

    def test_observations_sand_gravel(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY % ":SandGravel")
        assert results == [
            {"bound": PROBS.ExactBound, "measurement": Decimal('51010054000')},  # Obs 5
            {"bound": PROBS.ExactBound, "measurement": Decimal('56000000000')},  # Obs 4
            {"bound": PROBS.ExactBound, "measurement": Decimal('56660203000')},  # Obs 6
        ]


    def test_observations_aggregates(self, rdfox):
        results = rdfox.query_records(OBJECT_QUERY_TIME % ":Aggregates")
        assert results == [
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('150010054000')},  # Obs 2+5
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('152010054000')},  # Obs 3+5
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('155000000000')},  # Obs 2+4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2014, "measurement": Decimal('157000000000')},  # Obs 3+4
            {"bound": PROBS.ExactBound, "year": PROBS.TimePeriod_YearOf2018, "measurement": Decimal('126382754000')},  # Obs 1+6
        ]



# -*- coding: utf-8 -*-
"""Test effect of :objectEquivalentTo on Observations"""

from decimal import Decimal
import pytest

from probs_runner import PROBS

from utils import make_observation


############################################################
# Process and object definitions reused
############################################################

CAKE_ROLE_DEFS = r"""
:Cake a :Object ;
    rdfs:label "Cake" .
:Torta a :Object ;
    rdfs:label "Torta" .
:Baking a :Process ;
    rdfs:label "Baking" .
:CakeBaking a :Process ;
    rdfs:label "Baking" .
:Eating a :Process ;
    rdfs:label "Eating" .
:HavingTea a :Process ;
    rdfs:label "Eating" .
"""


############################################################
# Basic checks
############################################################

def test_basic_observations(probs_answer_query):
    """This is not a very good test, but checks everything is running ok."""
    facts = (
        CAKE_ROLE_DEFS
        + make_observation(obs_id="1", object=":Cake", measurement=15.4)
    )
    results = probs_answer_query(facts, r"""
        SELECT ?role ?measurement WHERE {
            :Observation1 :hasRole ?role ;
            :measurement ?measurement .
        }
    """)
    assert results == [
        {"role": PROBS.SoldProduction, "measurement": Decimal('15.4')}
    ]


def test_no_observation_without_equivalence(probs_answer_query):
    """When two objects not equivalent, observations should not be inferred across."""
    facts = (
        CAKE_ROLE_DEFS
        + make_observation(object=":Cake", measurement=15.4)
    )
    results = probs_answer_query(facts, r"""
        SELECT ?observation WHERE { ?observation :objectDefinedBy :Torta . }
    """)
    assert len(results) == 0


############################################################
# Check observations are propagated through equivalence
############################################################

@pytest.mark.parametrize("role", [":SoldProduction", ":Consumption", ":Import", ":Export"])
def test_object_only_roles_object_equivalence(probs_answer_query, role):
    facts = (
        CAKE_ROLE_DEFS
        + make_observation(role=role, object=":Cake", measurement=1)
        + ":Cake :objectEquivalentTo :Torta ."
    )
    results = probs_answer_query(facts, r"""
        SELECT ?observation WHERE { ?observation :objectDefinedBy :Torta . }
    """)
    assert len(results) == 1


def test_process_input_object_equivalence(probs_answer_query):
    facts = (
        CAKE_ROLE_DEFS
        + make_observation(role=":ProcessInput", object=":Cake", process=":Eating", measurement=1)
        + ":Cake :objectEquivalentTo :Torta ."
    )
    results = probs_answer_query(facts, r"""
        SELECT ?observation WHERE { ?observation :objectDefinedBy :Torta . }
    """)
    assert len(results) == 1


def test_process_output_object_equivalence(probs_answer_query):
    facts = (
        CAKE_ROLE_DEFS
        + make_observation(role=":ProcessOutput", object=":Cake", process=":CakeBaking", measurement=1)
        + ":Cake :objectEquivalentTo :Torta ."
    )
    results = probs_answer_query(facts, r"""
        SELECT ?observation WHERE { ?observation :objectDefinedBy :Torta . }
    """)
    assert len(results) == 1


# -*- coding: utf-8 -*-

import pytest
from random import randrange, random


from utils import make_observation


OBJECT_FACTS = r"""
:Object0 :objectComposedOf  :Object0.0 ,
                            :Object0.1 ,
                            :Object0.2 ,
                            :Object0.3 ,
                            :Object0.4 ,
                            :Object0.5 ,
                            :Object0.6 .
:Object0.0 :objectComposedOf    :Object0.0.0 ,
                                :Object0.0.1 ,
                                :Object0.0.2 ,
                                :Object0.0.3 ,
                                :Object0.0.4 ,
                                :Object0.0.5 ,
                                :Object0.0.6 .
:Object0.1 :objectComposedOf    :Object0.1.0 ,
                                :Object0.1.1 ,
                                :Object0.1.2 ,
                                :Object0.1.3 ,
                                :Object0.1.4 ,
                                :Object0.1.5 ,
                                :Object0.1.6 .
:Object0.2 :objectComposedOf    :Object0.2.0 ,
                                :Object0.2.1 ,
                                :Object0.2.2 ,
                                :Object0.2.3 ,
                                :Object0.2.4 ,
                                :Object0.2.5 ,
                                :Object0.2.6 .
:Object0.3 :objectComposedOf    :Object0.3.0 ,
                                :Object0.3.1 ,
                                :Object0.3.2 ,
                                :Object0.3.3 ,
                                :Object0.3.4 ,
                                :Object0.3.5 ,
                                :Object0.3.6 .
:Object0.4 :objectComposedOf    :Object0.4.0 ,
                                :Object0.4.1 ,
                                :Object0.4.2 ,
                                :Object0.4.3 ,
                                :Object0.4.4 ,
                                :Object0.4.5 ,
                                :Object0.4.6 .
:Object0.5 :objectComposedOf    :Object0.5.0 ,
                                :Object0.5.1 ,
                                :Object0.5.2 ,
                                :Object0.5.3 ,
                                :Object0.5.4 ,
                                :Object0.5.5 ,
                                :Object0.5.6 .
:Object0.6 :objectComposedOf    :Object0.6.0 ,
                                :Object0.6.1 ,
                                :Object0.6.2 ,
                                :Object0.6.3 ,
                                :Object0.6.4 ,
                                :Object0.6.5 ,
                                :Object0.6.6 .
"""

def test_obs_1_5x5_1eachsame_inferred_obs_37(probs_count_inferred_obs):
    """
    1 Object, 5 components each with 5 components, 1 Observations each, same probsObs
    Observations to be inferred: 37
    User time (seconds): ~1E0
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(5)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(5) for k in range(5)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(1)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 37

    # Full version: 5+32 (see below for some details)
    # WDF version:
    # Iteration 1:
    #   [:Object1.1 + ... + :Object1.5 -> :Object1]
    #     1  (D1.1+D1.2+D1.3+D1.4+D1.5)
    #   [:Object1.1.1 + ... + :Object1.1.5 -> :Object1.1]
    #     1  (D1.1.1+D1.1.2+D1.1.3+D1.1.4+D1.1.5)
    #   [:Object1.2.1 + ... + :Object1.2.5 -> :Object1.2]
    #     1  (D1.2.1+D1.2.2+D1.2.3+D1.2.4+D1.2.5)
    #   [:Object1.3.1 + ... + :Object1.3.5 -> :Object1.3]
    #     1  (D1.3.1+D1.3.2+D1.3.3+D1.3.4+D1.3.5)
    #   [:Object1.4.1 + ... + :Object1.4.5 -> :Object1.4]
    #     1  (D1.4.1+D1.4.2+D1.4.3+D1.4.4+D1.4.5)
    #   [:Object1.5.1 + ... + :Object1.5.5 -> :Object1.5]
    #     1  (D1.5.1+D1.5.2+D1.5.3+D1.5.4+D1.5.5)
    # Deleting duplicates:
    #     No duplicates
    # Iteration 2:
    #   [:Object1.1 + ... + :Object1.5 -> :Object1]
    #     32 (
    #     -1  ~~D1.1+D1.2+D1.3+D1.4+D1.5~~ already there
    #         D1.1+D1.2+D1.3+D1.4+I1.5,
    #         D1.1+D1.2+D1.3+I1.4+D1.5, D1.1+D1.2+D1.3+I1.4+I1.5,
    #         D1.1+D1.2+I1.3+D1.4+D1.5, D1.1+D1.2+I1.3+D1.4+I1.5, D1.1+D1.2+I1.3+I1.4+D1.5, D1.1+D1.2+I1.3+I1.4+I1.5,
    #         D1.1+I1.2+D1.3+D1.4+D1.5, D1.1+I1.2+D1.3+D1.4+I1.5, D1.1+I1.2+D1.3+I1.4+D1.5, D1.1+I1.2+D1.3+I1.4+I1.5, D1.1+I1.2+I1.3+D1.4+D1.5, D1.1+I1.2+I1.3+D1.4+I1.5, D1.1+I1.2+I1.3+I1.4+D1.5, D1.1+I1.2+I1.3+I1.4+I1.5,
    #         I1.1+D1.2+D1.3+D1.4+D1.5,
    #         I1.1+D1.2+D1.3+D1.4+I1.5,
    #         I1.1+D1.2+D1.3+I1.4+D1.5, I1.1+D1.2+D1.3+I1.4+I1.5,
    #         I1.1+D1.2+I1.3+D1.4+D1.5, I1.1+D1.2+I1.3+D1.4+I1.5, I1.1+D1.2+I1.3+I1.4+D1.5, I1.1+D1.2+I1.3+I1.4+I1.5,
    #         I1.1+I1.2+D1.3+D1.4+D1.5, I1.1+I1.2+D1.3+D1.4+I1.5, I1.1+I1.2+D1.3+I1.4+D1.5, I1.1+I1.2+D1.3+I1.4+I1.5, I1.1+I1.2+I1.3+D1.4+D1.5, I1.1+I1.2+I1.3+D1.4+I1.5, I1.1+I1.2+I1.3+I1.4+D1.5, I1.1+I1.2+I1.3+I1.4+I1.5,
    #        )
    #      or, equivalently
    #       DDDDD
    #       DDDDI
    #       DDDID DDDII
    #       DDIDD DDIDI DDIID DDIII
    #       DIDDD DIDDI DIDID DIDII DIIDD DIIDI DIIID DIIII
    #       IDDDD
    #       IDDDI
    #       IDDID IDDII
    #       IDIDD IDIDI IDIID IDIII
    #       IIDDD IIDDI IIDID IIDII IIIDD IIIDI IIIID IIIII
    # Deleting duplicates:
    #     No duplicates
    #

def test_obs_1_2x2_2eachsame_inferred_obs_44(probs_count_inferred_obs):
    """
    1 Object, 2 components each with 2 components, 2 Observations each, same probsObs.
    Observations to be inferred: 44
    User time (seconds): ~1E0
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(2)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(2) for k in range(2)]
    input_obs = "\n".join(
        make_observation(object=obj, obs_id=f"-{obj[7:]}{repeat}", measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 44

    # Full version: 4+4+4+36-4 (see below for some details)
    # WDF version:
    # Iteration 1:
    #   [:Object1.1 + :Object1.2 -> :Object1]
    #     4  (D1.1a+D1.2a, D1.1a+D1.2b, D1.1b+D1.2a, D1.1b+D1.2b)
    #   [:Object1.1.1 + :Object1.1.2 -> :Object1.1]
    #     4  (D1.1.1a+D1.1.2a, D1.1.1a+D1.1.2b, D1.1.1b+D1.1.2a, D1.1.1b+D1.1.2b)
    #   [:Object1.2.1 + :Object1.2.2 -> :Object1.2]
    #     4  (D1.2.1a+D1.2.2a, D1.2.1a+D1.1.2b, D1.2.1b+D1.2.2a, D1.2.1b+D1.2.2b)
    # Deleting duplicates:
    #     No duplicates
    # Iteration 2:
    #   [:Object1.1 + :Object1.2 -> :Object1]
    #     36 (
    #     -4  ~~D1.1a+D1.2a, D1.1a+D1.2b, D1.1b+D1.2a, D1.1b+D1.2b~~ already there
    #         D1.1a+I1.2aa, D1.1a+I1.2ab, D1.1a+I1.2ba, D1.1a+I1.2bb,
    #         D1.1b+I1.2aa, D1.1b+I1.2ab, D1.1b+I1.2ba, D1.1b+I1.2bb,
    #         I1.1aa+D1.2a, I1.1ab+D1.2a, I1.1ba+D1.2a, I1.1bb+D1.2a,
    #         I1.1aa+D1.2b, I1.1ab+D1.2b, I1.1ba+D1.2b, I1.1bb+D1.2b,
    #         I1.1aa+I1.2aa, I1.1aa+I1.2ab, I1.1aa+I1.2ba, I1.1aa+I1.2bb,
    #         I1.1ab+I1.2aa, I1.1ab+I1.2ab, I1.1ab+I1.2ba, I1.1ab+I1.2bb,
    #         I1.1ba+I1.2aa, I1.1ba+I1.2ab, I1.1ba+I1.2ba, I1.1ba+I1.2bb,
    #         I1.1bb+I1.2aa, I1.1bb+I1.2ab, I1.1bb+I1.2ba, I1.1bb+I1.2bb,
    #        )
    # Deleting duplicates:
    #     No duplicates
    #

def test_obs_1_3x3_2eachsame_inferred_obs_1024(probs_count_inferred_obs):
    """
    1 Object, 3 components each with 3 components, 2 Observations each, same probsObs.
    Observations to be inferred: 1024
    User time (seconds): ~1E0
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(3)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(3) for k in range(3)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 1024 # 8*3+8+1000-8

@pytest.mark.slow
def test_obs_1_4x4_2eachsame_inferred_obs_105040(probs_count_inferred_obs):
    """
    1 Object, 4 components each with 4 components, 2 Observations each, same probsObs.
    Observations to be inferred: 105040
    User time (seconds): ~1E1
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(4)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(4) for k in range(4)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 105040 # 16*4+16+104976-16

@pytest.mark.slow
def test_obs_1_4x5_2eachsame_inferred_obs_1336464(probs_count_inferred_obs):
    """
    1 Object, 4 components each with 5 components, 2 Observations each, same probsObs.
    Observations to be inferred: 1336464
    User time (seconds): ~1E2
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(4)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(4) for k in range(5)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 1336464 # 2^5*4+2^4+1336336-16

@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_1_4x6_2eachsame_inferred_obs_18974992(probs_count_inferred_obs):
    """
    1 Object, 4 components each with 6 components, 2 Observations each, same probsObs.
    Observations to be inferred: 18974992
    User time (seconds): ~1E3
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(4)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(4) for k in range(6)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 18974992 # 2^6*4+2^4+(2^6+2)^4-16=256+18974736

@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_1_5x5_2eachsame_inferred_obs_45435584(probs_count_inferred_obs):
    """
    1 Object, 5 components each with 5 components, 2 Observations each, same
    probsObs.
    Observations to be inferred: 45435584
    
    CANNOT WORK! It should create 45'435'584 of new Observations!!
    Assuming 1 byte per character and that each observation contains around 1000
    characters (I have checked this in the output of PRODCOM), only to store
    them we would need more than 40 GB to store them (in RAM, while processing
    it, they would be much more)
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(5)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(5) for k in range(5)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(2)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 45435584 # 32*5+32+45435424-32

# @pytest.mark.slow
def test_obs_1_3x3_3eachsame_inferred_obs_27081(probs_count_inferred_obs):
    """
    1 Object, 3 components each with 3 components, 3 Observations each, same probsObs.
    Observations to be inferred: 27081
    User time (seconds): ~1E0
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(3)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(3) for k in range(3)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(3)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 27081 # 27*3+27+27000-27

@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_1_4x4_3eachsame_inferred_obs_49787460(probs_count_inferred_obs):
    """
    1 Object, 4 components each with 4 components, 3 Observations each, same
    probsObs.
    Observations to be inferred: 49787460

    CANNOT WORK! It should create 49'787'460 of new Observations!!
    """

    obj_level1 = [f":Object{i}.{j}" for i in range(1) for j in range(4)]
    obj_level2 = [f":Object{i}.{j}.{k}" for i in range(1) for j in range(4) for k in range(4)]
    input_obs = "\n".join(
        make_observation(object=obj, measurement=round(random(), 6))
        for obj_list in [obj_level1, obj_level2]
        for obj in obj_list
        for _ in range(3)
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs) == 49787460 # 81*4+81+49787136-81



# -*- coding: utf-8 -*-
"""Test case of an object composed of objects, composed of objects, ... with
observations only in the bottom level of this hierarchy."""

import pytest
from random import random
from utils import make_observation

# Min: (lower bound) Nobs * Ncomp_{n} * Ncomp_{n-1} * ... * Ncomp_{0}
# Max: (upper bound) (...(Nobs ^ Ncomp_{n}) ^ Ncomp_{n-1}) ^ ...) ^ Ncomp_{0}

levels = 6

OBJECT_FACTS = \
    '\n'.join([f":Object0 :objectComposedOf :Object0.{i} ." for i in range(levels)]) + \
        '\n'.join([f":Object0.{i} :objectComposedOf :Object0.{i}.{j} ." for i in range(levels) for j in range(levels)]) + \
            '\n'.join([f":Object0.{i}.{j} :objectComposedOf :Object0.{i}.{j}.{k} ." for i in range(levels) for j in range(levels) for k in range(levels)]) + \
                '\n'.join([f":Object0.{i}.{j}.{k} :objectComposedOf :Object0.{i}.{j}.{k}.{l} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels)]) + \
                    '\n'.join([f":Object0.{i}.{j}.{k}.{l} :objectComposedOf :Object0.{i}.{j}.{k}.{l}.{m} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels) for m in range(levels)]) + \
                        '\n'.join([f":Object0.{i}.{j}.{k}.{l}.{m} :objectComposedOf :Object0.{i}.{j}.{k}.{l}.{m}.{n} ." for i in range(levels) for j in range(levels) for k in range(levels) for l in range(levels) for m in range(levels) for n in range(levels)])

def test_obs_obj1_comp1x2x3_obs2_min_inferred_obs_12(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 12
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(1) for j in range(2) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*3*2*1

def test_obs_obj1_comp1x2x3_obs2_max_inferred_obs_64(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 64
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(1) for j in range(2) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == pow(pow(pow(2,3),2),1)

def test_obs_obj1_comp1x2x3_obs3_min_inferred_obs_18(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components
    3 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 18
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(1) for j in range(2) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "abc"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 3*3*2*1

def test_obs_obj1_comp1x2x3_obs3_max_inferred_obs_729(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components
    3 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 729
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(1) for j in range(2) for k in range(3)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "abc"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == pow(pow(pow(3,3),2),1)

def test_obs_obj1_comp1x2x3x4_obs2_min_inferred_obs_48(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 48
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(1) for j in range(2) for k in range(3) for l in range(4)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*4*3*2*1

@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_obj1_comp1x2x3x4_obs2_max_inferred_obs_16777216(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 16'777'216
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(1) for j in range(2) for k in range(3) for l in range(4)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") ==  pow(pow(pow(pow(2,4),3),2),1)

def test_obs_obj1_comp1x2x3x4x5_obs2_min_inferred_obs_240(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 240
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*5*4*3*2*1

def test_obs_obj1_comp1x2x3x4x5x6_obs2_min_inferred_obs_1440(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components each with 6 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 1440
    User time (seconds): ~1E0
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5) for n in range(6)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*6*5*4*3*2*1

@pytest.mark.slow
def test_obs_obj1_comp1x2x3x4x5x6_obs3_min_inferred_obs_2160(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components each with 6 components
    3 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 2160
    User time (seconds): ~1E1
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5) for n in range(6)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "abc"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 3*6*5*4*3*2*1

@pytest.mark.slow
def test_obs_obj1_comp1x2x3x4x5x6_obs4_min_inferred_obs_2880(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components each with 6 components
    4 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 2880
    User time (seconds): ~1E1
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5) for n in range(6)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "abcd"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 4*6*5*4*3*2*1

@pytest.mark.slow
def test_obs_obj1_comp1x2x3x4x5x6_obs5_min_inferred_obs_3600(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components each with 6 components
    5 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 3600
    User time (seconds): ~1E1
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5) for n in range(6)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "abcde"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 5*6*5*4*3*2*1

@pytest.mark.slow
def test_obs_obj1_comp1x2x3x4x5x6_obs6_min_inferred_obs_4320(probs_count_inferred_obs):
    """
    1 Object
    1 components each with 2 components each with 3 components each with 4 components each with 5 components each with 6 components
    5 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 4320
    User time (seconds): ~1E1
    """

    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(1) for j in range(2) for k in range(3) for l in range(4) for m in range(5) for n in range(6)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "abcdef"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 6*6*5*4*3*2*1

def test_obs_obj1_comp2x2x2_obs2_max_inferred_obs_256(probs_count_inferred_obs):
    """
    1 Object
    2 components each with 2 components each with 2 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 256
    User time (seconds): ~1E0
    """

    components = 2
    objects = [
        f":Object0.{i}.{j}.{k}" for i in range(components) for j in range(components) for k in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") ==  pow(pow(pow(2,components),components),components)

@pytest.mark.skip(reason="our computers cannot process this, it is too big")
def test_obs_obj1_comp3x3x3_obs2_max_inferred_obs_134217728(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 134'217'728
    """

    components = 3
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == pow(pow(pow(2,components),components),components)

@pytest.mark.slow
def test_obs_obj1_comp2x2x2x2_obs2_max_inferred_obs_65536(probs_count_inferred_obs):
    """
    1 Object
    2 components each with 2 components each with 2 components each with 2 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 65536
    User time (seconds): ~1E1
    """

    components = 2
    objects = [
        f":Object0.{i}.{j}.{k}.{l}" for i in range(components) for j in range(components) for k in range(components) for l in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R-same-for-all",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") ==  pow(pow(pow(pow(2,components),components),components),components)

def test_obs_obj1_comp2x2x2x2x2_obs2_min_inferred_obs_64(probs_count_inferred_obs):
    """
    1 Object
    2 components each with 2 components each with 2 components each with 2 components each with 2 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 64
    User time (seconds): ~1E0
    """

    components = 2
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)

def test_obs_obj1_comp3x3x3x3x3_obs2_min_inferred_obs_486(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 486
    User time (seconds): ~1E0
    """

    components = 3
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)

@pytest.mark.slow
def test_obs_obj1_comp4x4x4x4x4_obs2_min_inferred_obs_2048(probs_count_inferred_obs):
    """
    1 Object
    4 components each with 4 components each with 4 components each with 4 components each with 4 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 2048
    User time (seconds): ~1E1
    """

    components = 4
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)

@pytest.mark.slow
def test_obs_obj1_comp5x5x5x5x5_obs2_min_inferred_obs_6250(probs_count_inferred_obs):
    """
    1 Object
    5 components each with 5 components each with 5 components each with 5 components each with 5 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 6250
    User time (seconds): ~1E1
    """

    components = 5
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,5)

def test_obs_obj1_comp2x2x2x2x2x2_obs2_min_inferred_obs_128(probs_count_inferred_obs):
    """
    1 Object
    2 components each with 2 components each with 2 components each with 2 components each with 2 components each with 2 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 128
    User time (seconds): ~1E0
    """

    components = 2
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components) for n in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,levels)

def test_obs_obj1_comp3x3x3x3x3x3_obs2_min_inferred_obs_1458(probs_count_inferred_obs):
    """
    1 Object
    3 components each with 3 components each with 3 components each with 3 components each with 3 components each with 3 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 1458
    User time (seconds): ~1E0
    """

    components = 3
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components) for n in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,levels)

@pytest.mark.slow
def test_obs_obj1_comp4x4x4x4x4x4_obs2_min_inferred_obs_8192(probs_count_inferred_obs):
    """
    1 Object
    4 components each with 4 components each with 4 components each with 4 components each with 4 components each with 4 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 8192
    User time (seconds): ~1E1
    """

    components = 4
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components) for n in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,levels)

@pytest.mark.slow
def test_obs_obj1_comp5x5x5x5x5x5_obs2_min_inferred_obs_31250(probs_count_inferred_obs):
    """
    1 Object
    5 components each with 5 components each with 5 components each with 5 components each with 5 components each with 5 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 31250
    User time (seconds): ~1E2
    """

    components = 5
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components) for n in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,levels)

@pytest.mark.slow
def test_obs_obj1_comp6x6x6x6x6x6_obs2_min_inferred_obs_93312(probs_count_inferred_obs):
    """
    1 Object
    6 components each with 6 components each with 6 components each with 6 components each with 6 components each with 6 components
    2 Observations for each object at the bottom
    lower bound (non-compatible observations)
    Observations to be inferred: 93312
    User time (seconds): ~1E2
    """

    components = 6
    objects = [
        f":Object0.{i}.{j}.{k}.{l}.{m}.{n}" for i in range(components) for j in range(components) for k in range(components) for l in range(components) for m in range(components) for n in range(components)
    ]
    input_obs = "\n".join(
        make_observation(
            object=obj,
            obs_id=f"-{obj[7:]}{repeat}",
            measurement=round(random(), 6),
            region=f":R{obj[7:]}{repeat}",
        )
        for obj in objects
        for repeat in "ab"
    )
    assert probs_count_inferred_obs(OBJECT_FACTS + input_obs, ":Object0") == 2*pow(components,levels)
